package de.swe2.bestellverwaltung.domain;

import static de.swe2.util.EjbConstants.ERSTE_VERSION;
import static de.swe2.util.EjbConstants.KEINE_ID;
import static javax.persistence.CascadeType.PERSIST;
import static javax.persistence.EnumType.STRING;
import static javax.persistence.TemporalType.TIMESTAMP;
import static javax.xml.bind.annotation.XmlAccessType.FIELD;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.Enumerated;
import javax.persistence.GeneratedValue;
import javax.persistence.Id;
import javax.persistence.ManyToMany;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.OrderBy;
import javax.persistence.PostPersist;
import javax.persistence.PostUpdate;
import javax.persistence.PrePersist;
import javax.persistence.PreUpdate;
import javax.persistence.SequenceGenerator;
import javax.persistence.Table;
import javax.persistence.Temporal;
import javax.persistence.Transient;
import javax.persistence.Version;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlTransient;

import org.hibernate.validator.constraints.NotEmpty;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


@Entity
@Table(name = "lieferung")
@NamedQueries({
	@NamedQuery(name  = Lieferung.FIND_LIEFERUNGEN_BY_LIEFERNR_FETCH_BESTELLUNGEN,
                query = "SELECT l"
                	    + " FROM Lieferung l LEFT JOIN FETCH l.bestellungen"
			            + " WHERE l.lieferNr LIKE :" + Lieferung.PARAM_LIEFER_NR)
})
@XmlRootElement
@XmlAccessorType(FIELD)
public class Lieferung implements java.io.Serializable {
	private static final long serialVersionUID = 7560752199018702446L;
	protected static final Logger LOGGER = LoggerFactory.getLogger(Lieferung.class);
	
	private static final int LIEFERNR_LENGTH_MAX = 12;
	
	private static final String PREFIX = "Lieferung.";
	public static final String FIND_LIEFERUNGEN_BY_LIEFERNR_FETCH_BESTELLUNGEN = PREFIX + "findLieferungenByLiefernrFetchBestellungen";
	public static final String PARAM_LIEFER_NR = "lieferNr";
	
	@XmlEnum
	public enum TransportArt {
		STRASSE,
		SCHIENE,
		LUFT,
		WASSER;
	}
	
	@Id
	@GeneratedValue(generator = "lieferung_sequence_name")
	@SequenceGenerator(name = "lieferung_sequence_name", sequenceName = "lieferung_l_id_seq", allocationSize = 1)
	@Column(name = "l_id", nullable = false)
	@XmlAttribute(name = "id", required = true)
	private Long id = KEINE_ID;

	@Version
	@XmlTransient
	private int version = ERSTE_VERSION;

	@Column(name = "liefernr", length = LIEFERNR_LENGTH_MAX, nullable = false, unique = true)
	@NotNull(message = "{bestellverwaltung.lieferung.lieferNr.notNull}")
	@Size(max = LIEFERNR_LENGTH_MAX, message = "{bestellverwaltung.lieferung.lieferNr.length}")
	private String lieferNr;
	
	@Column(name = "transport_art")
	@Enumerated(STRING)
	private TransportArt transportArt;

	@ManyToMany(mappedBy = "lieferungen", cascade = PERSIST)
	@OrderBy("id ASC")
	@NotEmpty(message = "{bestellverwaltung.lieferung.bestellungen.notEmpty}")
	@Valid
	@XmlTransient
	private List<Bestellung> bestellungen;
	
	@Transient
	@XmlElementWrapper(name = "bestellungen")
	@XmlElement(name = "bestellung")
	private List<URI> bestellungenKeys;
	
	@Column(nullable = false)
	@Temporal(TIMESTAMP)
	private Date erzeugt = null;

	@Column(nullable = false)
	@Temporal(TIMESTAMP)
	private Date aktualisiert = null;
	
	public Lieferung() {
		super();
	}
	
	public Lieferung(List<Bestellung> bestellungen) {
		super();
		this.bestellungen = bestellungen;
	}
	
	@SuppressWarnings("unused")
	@PrePersist
	private void prePersist() {
		erzeugt = new Date();
		preUpdate();
	}
	
	@SuppressWarnings("unused")
	@PostPersist
	private void logDbId() {
		LOGGER.debug("Neue Lieferung mit ID={}", id);
	}

	@PreUpdate
	private void preUpdate() {
		aktualisiert = new Date();
	}

	@SuppressWarnings("unused")
	@PostUpdate
	private void logNeueVersion() {
		LOGGER.trace("Lieferung ID={} geaendert: neue Version={}", id, version);
	}

	public Long getId() {
		return id;
	}
	public void setId(Long id) {
		this.id = id;
	}

	public int getVersion() {
		return version;
	}
	public void setVersion(int version) {
		this.version = version;
	}

	public String getLieferNr() {
		return lieferNr;
	}
	public void setLieferNr(String lieferNr) {
		this.lieferNr = lieferNr;
	}

	public TransportArt getTransportArt() {
		return transportArt;
	}
	public void setTransportArt(TransportArt transportArt) {
		this.transportArt = transportArt;
	}

	public List<Bestellung> getBestellungen() {
		if (bestellungen == null) {
			return null;
		}
		
		return Collections.unmodifiableList(bestellungen);
	}
	public void setBestellungen(List<Bestellung> bestellungen) {
		if (this.bestellungen == null) {
			this.bestellungen = bestellungen;
			return;
		}
		
		// Wiederverwendung der vorhandenen Collection
		this.bestellungen.clear();
		if (bestellungen != null) {
			this.bestellungen.addAll(bestellungen);
		}
	}
	
	public void addBestellung(Bestellung bestellung) {
		if (bestellungen == null) {
			bestellungen = new ArrayList<Bestellung>();
		}
		bestellungen.add(bestellung);
	}

	public List<URI> getBestellungenKeys() {
		return bestellungenKeys;
	}
	public void setBestellungenKeys(List<URI> bestellungenKeys) {
		this.bestellungenKeys = bestellungenKeys;
	}

	public Date getErzeugt() {
		return erzeugt == null ? null : (Date) erzeugt.clone();
	}

	public void setErzeugt(Date erzeugt) {
		this.erzeugt = erzeugt == null ? null : (Date) erzeugt.clone();
	}

	public Date getAktualisiert() {
		return aktualisiert == null ? null : (Date) aktualisiert.clone();
	}

	public void setAktualisiert(Date aktualisiert) {
		this.aktualisiert = aktualisiert == null ? null : (Date) aktualisiert.clone();
	}
	
	@Override
	public String toString() {
		return "Lieferung [id=" + id + ", version=" + version
		       + ", lieferNr=" + lieferNr
		       + ", transportArt=" + transportArt
		       + ", erzeugt=" + erzeugt
		       + ", aktualisiert=" + aktualisiert + ']';
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((lieferNr == null) ? 0 : lieferNr.hashCode());
		result = prime * result + version;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final Lieferung other = (Lieferung) obj;
		if (lieferNr == null) {
			if (other.lieferNr != null) {
				return false;
			}
		}
		else if (!lieferNr.equals(other.lieferNr)) {
			return false;
		}
		if (version != other.version) {
			return false;
		}
		return true;
	}
}
