package de.swe2.bestellverwaltung.service;


import static de.swe2.util.AbstractDao.QueryParameter.with;
import static de.swe2.util.EjbConstants.KEINE_ID;
import static de.swe2.util.EjbConstants.ROLLE_KUNDE;

import java.io.Serializable;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import javax.annotation.PreDestroy;
import javax.annotation.security.RolesAllowed;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.validation.ConstraintViolation;
import javax.validation.Validator;
import javax.validation.groups.Default;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.swe2.artikelverwaltung.domain.Artikel;
import de.swe2.bestellverwaltung.domain.Bestellposition;
import de.swe2.bestellverwaltung.domain.Bestellung;
import de.swe2.bestellverwaltung.domain.Lieferung;
import de.swe2.bestellverwaltung.domain.PlzAnzahlBestellungen;
import de.swe2.kundenverwaltung.domain.AbstractKunde;
import de.swe2.kundenverwaltung.service.InvalidKundeIdException;
import de.swe2.kundenverwaltung.service.Kundenverwaltung;
import de.swe2.util.IdGroup;
import de.swe2.util.NotFoundException;
import de.swe2.util.ValidationService;


@Stateless
public class Bestellverwaltung implements Serializable {
	private static final long serialVersionUID = 3428176213374643888L;
	private static final Logger LOGGER = LoggerFactory.getLogger(Bestellverwaltung.class);
	
	@PersistenceContext
	@SuppressWarnings("unused")
	private transient EntityManager em;
	
	@EJB
	private ValidationService validationService;
	
	@EJB
	private BestellverwaltungDao dao;
	
	@EJB
	private Kundenverwaltung kv;
	
	@SuppressWarnings("unused")
	@PreDestroy
	private void preDestroy() {
		LOGGER.debug("Session Bean {} wird geloescht", Bestellverwaltung.class.getSimpleName());
	}
	
	/**
	 */
	public Bestellung findBestellungById(Long id, Locale locale) throws NotFoundException, InvalidBestellungIdException {
		validateBestellungId(id, locale);
		final Bestellung bestellung = dao.find(Bestellung.class, id);
		return bestellung;
	}
	
	private void validateBestellungId(Long id, Locale locale) throws InvalidBestellungIdException {
		LOGGER.debug("BEGINN validateArtikelId: id={}", id);
		final Validator validator = validationService.getValidator(locale);
		Set<ConstraintViolation<Bestellung>> violations = validator.validateValue(Bestellung.class, "id", id, IdGroup.class);	
		
		if (!violations.isEmpty()) {
			throw new InvalidBestellungIdException(id, violations);
		}		
		LOGGER.debug("ENDE validateArtikelId");
	}

	/**
	 */
	public Bestellung findBestellungByIdMitLieferungen(Long id, Locale locale) throws NotFoundException, InvalidBestellungIdException {
		validateBestellungId(id, locale);
		final Bestellung bestellung = dao.findSingle(Bestellung.FIND_BESTELLUNG_BY_ID_FETCH_LIEFERUNGEN,
				                                     with(Bestellung.PARAM_ID, id).parameters(),
				                                     Bestellung.class);
		return bestellung;
	}
	
	/**
	 */
	public List<Bestellung> findBestellungenByKunde(AbstractKunde kunde) throws NotFoundException {
		final List<Bestellung> bestellungen = dao.find(Bestellung.FIND_BESTELLUNGEN_BY_KUNDEID,
                                                       with(Bestellung.PARAM_KUNDEID, kunde.getId()).parameters(),
                                                       Bestellung.class);
		return bestellungen;
	}

	
	/**
	 */
	public AbstractKunde findKundeById(Long id, Locale locale) throws NotFoundException, InvalidBestellungIdException {
		validateBestellungId(id, locale);
		final AbstractKunde kunde = dao.findSingle(Bestellung.FIND_KUNDE_BY_ID,
                                                   with(Bestellung.PARAM_ID, id).parameters(),
                                                   AbstractKunde.class);
		return kunde;
	}
	
	/**
	 */
	public List<PlzAnzahlBestellungen> findPlzAnzahlBestellungen() throws NotFoundException {
		final List<PlzAnzahlBestellungen> result = dao.findPlzAnzahlBestellungen();
		return result;
	}


	/**
	 */
	public List<Bestellung> findBestellungenMitLieferungenByKunde(AbstractKunde kunde) throws NotFoundException {
		final List<Bestellung> bestellungen = dao.find(Bestellung.FIND_BESTELLUNGEN_BY_KUNDEID_FETCH_LIEFERUNGEN,
                                                       with(Bestellung.PARAM_KUNDEID, kunde.getId()).parameters(),
                                                       Bestellung.class);
		return bestellungen;
	}

	/**
	 */
	public long findAnzBestellungenByPlzV1(int min, int max) throws NotFoundException {
		final Long anzahl = dao.findSingle(Bestellung.FIND_ANZ_BESTELLUNGEN_BY_PLZ_V1,
                                           with(Bestellung.PARAM_PLZ_MIN, min).and(Bestellung.PARAM_PLZ_MAX, max).parameters(),
                                           Long.class);
		return anzahl.longValue();
	}
	
	/**
	 */
	public long findAnzBestellungenByPlzV2(int min, int max) throws NotFoundException {
		final Long anzahl = dao.findSingle(Bestellung.FIND_ANZ_BESTELLUNGEN_BY_PLZ_V2,
                                           with(Bestellung.PARAM_PLZ_MIN, min).and(Bestellung.PARAM_PLZ_MAX, max).parameters(),
                                           Long.class);
		return anzahl.longValue();
	}

	
	/**
	 * Die Bestellung ist dem Kunden noch nicht hinzugefuegt, weil dieser
	 * sonst evtl. in einer Transaktion modifiziert wurde.
	 * Beim naechsten DB-Zugriff (auch lesend!) wuerde der EntityManager sonst
	 * erstmal versuchen den Kunden-Datensatz in der DB zu modifizieren.
	 * Dann wuerde aber der Kunde mit einer *transienten* Bestellung modifiziert werden,
	 * was zwangslaeufig zu einem inkonsistenten Mix aus einem "managed" Objekt (hier: "kunde")
	 * und einem referenzierten, aber transienten Objekt (hier: "bestellung") fuehrt!
	 */
	@RolesAllowed(ROLLE_KUNDE)
	public Bestellung createBestellung(Bestellung bestellung, AbstractKunde kunde, Locale locale, boolean uiValidated)
	       throws NotFoundException, BestellungValidationException, BestellungDuplikatException, InvalidKundeIdException {
		
		if (!uiValidated) {
			final Validator validator = validationService.getValidator(locale);
			// Nicht alle Attribute validieren: der Kunde kann noch nicht vorhanden sein
			final Set<ConstraintViolation<Bestellung>> violations = validator.validateValue(Bestellung.class, "bestellpositionen", bestellung.getBestellpositionen(), Default.class);
			if (violations != null && !violations.isEmpty()) {
				throw new BestellungValidationException(bestellung, violations);
			}
		}
		
		// Pruefung bzgl. Duplikat
		kunde = kv.findKundeMitBestellungenById(kunde.getId(), locale);  // kunde ist jetzt auch "managed"
		final List<Bestellung> bestellungen = kunde.getBestellungen();
		for (Bestellung b : bestellungen) {
			if (!bestellung.equals(b)) {
				continue;
			}
					
			throw new BestellungDuplikatException(kunde.getId(), bestellung.getErzeugt());
		}
		
		// Den persistenten Kunden mit der transienten Bestellung verknuepfen:
		// Der Kunde ist bereits "managed" (siehe oben), deshalb kein "reload" notwendig
		bestellung.setKunde(kunde);
		kunde.addBestellung(bestellung);
		
		// Vor dem Abspeichern IDs zuruecksetzen:
		// IDs koennten z.B. 0 sein, wenn sie durch einen Web Service uebertragen wurden
		bestellung.setId(KEINE_ID);
		for (Bestellposition bp : bestellung.getBestellpositionen()) {
			bp.setId(KEINE_ID);
		}
		bestellung = dao.create(bestellung);
		
		return bestellung;
	}
	
	/**
	 */
	public List<Artikel> ladenhueter(int anzahl) throws NotFoundException {
		return dao.find(Bestellposition.FIND_LADENHUETER, anzahl, Artikel.class);
	}
	
	/**
	 */
	public List<Lieferung> findLieferungen(String nr) throws NotFoundException {
		final List<Lieferung> lieferungen = dao.find(Lieferung.FIND_LIEFERUNGEN_BY_LIEFERNR_FETCH_BESTELLUNGEN,
                                                     with(Lieferung.PARAM_LIEFER_NR, nr).parameters(),
                                                     Lieferung.class);
		return lieferungen;
	}
	
	/**
	 */
	public Lieferung createLieferung(Lieferung lieferung) {
		if (lieferung == null) {
			return null;
		}
		
		lieferung.setId(KEINE_ID);
		lieferung = dao.create(lieferung);
		return lieferung;
	}
}
