package de.swe2.kundenverwaltung.service;

import static de.swe2.util.AbstractDao.QueryParameter.with;
import static de.swe2.util.EjbConstants.KEINE_ID;
import static de.swe2.util.EjbConstants.ROLLE_ADMIN;
import static de.swe2.util.EjbConstants.ROLLE_MITARBEITER;
import static de.swe2.util.jboss.PasswordUtil.verschluesseln;
import static de.swe2.util.jboss.RolesUtil.refreshRoles;
import static javax.ejb.TransactionAttributeType.SUPPORTS;

import java.io.File;
import java.io.Serializable;
import java.security.Principal;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.concurrent.Future;

import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.annotation.security.RolesAllowed;
import javax.ejb.AsyncResult;
import javax.ejb.Asynchronous;
import javax.ejb.EJB;
import javax.ejb.SessionContext;
import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.mail.Message.RecipientType;
import javax.mail.MessagingException;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.security.auth.Subject;
import javax.security.jacc.PolicyContext;
import javax.security.jacc.PolicyContextException;
import javax.validation.ConstraintViolation;
import javax.validation.Validator;
import javax.validation.groups.Default;

import org.jboss.security.SimpleGroup;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.swe2.kundenverwaltung.domain.AbstractKunde;
import de.swe2.kundenverwaltung.domain.AbstractKunde.PasswordGroup;
import de.swe2.kundenverwaltung.domain.Adresse;
import de.swe2.kundenverwaltung.domain.Betreuer;
import de.swe2.kundenverwaltung.domain.Privatkunde;
import de.swe2.kundenverwaltung.domain.Wartungsvertrag;
import de.swe2.util.ConcurrentDeletedException;
import de.swe2.util.ConcurrentUpdatedException;
import de.swe2.util.IdGroup;
import de.swe2.util.NotFoundException;
import de.swe2.util.ValidationService;


/**
 * Anwendungskern mit Plausibilit&auml;tspr&uuml;fungen und Aufruf der DAOs
 */
@Stateless
//JBoss-spezifische Annotation fuer Security-Domain (statt jboss-app.xml oder jboss.xml)
//@SecurityDomain("swe2")
public class Kundenverwaltung implements Serializable {
	private static final long serialVersionUID = 2360123134858497516L;
	private static final Logger LOGGER = LoggerFactory.getLogger(Kundenverwaltung.class);
	private static final String NEWLINE = System.getProperty("line.separator");
	
	@SuppressWarnings("unused")
	@PersistenceContext
	private transient EntityManager em;
	
	@EJB
	private ValidationService validationService;
	
	@EJB
	private KundenverwaltungDao dao;
	
	@Resource(name = "maxKategorie")         // ggf. ueberschreiben in ejb-jar.xml
	private final Short maxKategorie = Short.valueOf((short) 0);
	
	@Resource(name = "maxRabatt")         // ggf. ueberschreiben in ejb-jar.xml
	private final Float maxRabatt = Float.valueOf(0);
	
	@Resource(mappedName = "java:Mail")    // mappedName notwendig fuer JBoss
	private transient javax.mail.Session mailer;
	
	@Resource
	private transient SessionContext sessionCtx;

	@SuppressWarnings("unused")
	@PreDestroy
	private void preDestroy() {
		LOGGER.debug("Session Bean {} wird geloescht", Kundenverwaltung.class.getSimpleName());
	}

	/**
	 */
	public short getMaxKategorie() {
		final short result = maxKategorie.shortValue();
		return result;
	}
	
	/**
	 */
	public float getMaxRabatt() {
		final float result = maxRabatt.floatValue();
		return result;
	}

	/**
	 */
	public List<AbstractKunde> findAllKunden() throws NotFoundException {
		final List<AbstractKunde> kunden = dao.find(AbstractKunde.FIND_KUNDEN, AbstractKunde.class);
		return kunden;
	}
	
	/**
	 */
	public List<AbstractKunde> findAllKundenMitBestellungen() throws NotFoundException {
		final List<AbstractKunde> kunden = dao.find(AbstractKunde.FIND_KUNDEN_FETCH_BESTELLUNGEN, AbstractKunde.class);
		return kunden;
	}
	
	/**
	 */
	public List<AbstractKunde> findAllKundenOrderById() throws NotFoundException {
		final List<AbstractKunde> kunden = dao.find(AbstractKunde.FIND_KUNDEN_ORDER_BY_ID, AbstractKunde.class);
		return kunden;
	}

	/**
	 */
	public List<AbstractKunde> findKundenByNachname(String nachname, Locale locale) throws NotFoundException, InvalidNachnameException {
		validateNachname(nachname, locale);
		
		final List<AbstractKunde> kunden = dao.find(AbstractKunde.FIND_KUNDEN_BY_NACHNAME,
				                                    with(AbstractKunde.PARAM_KUNDE_NACHNAME, nachname).parameters(),
				                                    AbstractKunde.class);
		return kunden;
	}

	/**
	 */
	public List<String> findNachnamenByPrefix(String nachnamePrefix) throws NotFoundException {
		final List<String> nachnamen = dao.find(AbstractKunde.FIND_NACHNAMEN_BY_PREFIX,
		                                        with(AbstractKunde.PARAM_KUNDE_NACHNAME_PREFIX, nachnamePrefix + '%').parameters(),
		                                        String.class);
		return nachnamen;
	}

	
	/**
	 */
	public List<AbstractKunde> findKundenMitBestellungen(String nachname, Locale locale)
	                           throws NotFoundException, InvalidNachnameException {
		validateNachname(nachname, locale);
		
		final List<AbstractKunde> kunden = dao.find(AbstractKunde.FIND_KUNDEN_BY_NACHNAME_FETCH_BESTELLUNGEN,
                                                    with(AbstractKunde.PARAM_KUNDE_NACHNAME, nachname).parameters(),
                                                    AbstractKunde.class);
		return kunden;
	}
	
	/**
	 */
	public List<AbstractKunde> findKundenMitBestellungenUndLieferungen(String nachname, Locale locale)
	                           throws NotFoundException, InvalidNachnameException {
		validateNachname(nachname, locale);
		final List<AbstractKunde> kunden = dao.findKundenMitBestellungenUndLieferungen(nachname);
		return kunden;
	}
	
	/**
	 */
	public AbstractKunde findKundeMitWartungsvertraegenById(Long id, Locale locale) throws NotFoundException, InvalidKundeIdException {
		validateKundeId(id, locale);
		final AbstractKunde kunde = dao.findSingle(AbstractKunde.FIND_KUNDE_BY_ID_FETCH_WARTUNGSVERTRAEGE,
                                                   with(AbstractKunde.PARAM_KUNDE_ID, id).parameters(),
                                                   AbstractKunde.class);
		return kunde;
	}

	/**
	 */
	public AbstractKunde findKundeById(Long id, Locale locale) throws NotFoundException, InvalidKundeIdException {
		validateKundeId(id, locale);

		final AbstractKunde kunde = dao.find(AbstractKunde.class, id);
		return kunde;
	}

	/**
	 */
	public AbstractKunde findKundeMitBestellungenById(Long id, Locale locale) throws NotFoundException, InvalidKundeIdException {
		validateKundeId(id, locale);
		
		final AbstractKunde kunde = dao.findSingle(AbstractKunde.FIND_KUNDE_BY_ID_FETCH_BESTELLUNGEN,
                                                   with(AbstractKunde.PARAM_KUNDE_ID, id).parameters(),
                                                   AbstractKunde.class);
		return kunde;
	}
	
	/**
	 */
	public AbstractKunde findKundeByEmail(String email, Locale locale) throws NotFoundException, InvalidEmailException {
		validateEmail(email, locale);
		
		final AbstractKunde kunde = dao.findSingle(AbstractKunde.FIND_KUNDE_BY_EMAIL,
                                                   with(AbstractKunde.PARAM_KUNDE_EMAIL, email).parameters(),
                                                   AbstractKunde.class);
		return kunde;
	}
	
	private void validateEmail(String email, Locale locale) throws InvalidEmailException {
		LOGGER.debug("BEGINN validateEmail: email={}", email);
		final Validator validator = validationService.getValidator(locale);
		Set<ConstraintViolation<AbstractKunde>> violations = validator.validateValue(AbstractKunde.class, "email", email);	
		
		if (!violations.isEmpty()) {
			throw new InvalidEmailException(email, violations);
		}
		LOGGER.debug("ENDE validateEmail");
	}
	
	/**
	 */
	public List<AbstractKunde> findKundenByPLZ(String plz, Locale locale) throws NotFoundException, InvalidAdressePlzException {
		validateAdressePlz(plz, locale);
		
		final List<AbstractKunde> kunden = dao.find(AbstractKunde.FIND_KUNDEN_BY_PLZ,
                                                    with(AbstractKunde.PARAM_KUNDE_ADRESSE_PLZ, plz).parameters(),
                                                    AbstractKunde.class);
		return kunden;
	}
	
	private void validateAdressePlz(String plz, Locale locale) throws InvalidAdressePlzException {
		LOGGER.debug("BEGINN validateAdressePlz: plz={}", plz);
		final Validator validator = validationService.getValidator(locale);
		Set<ConstraintViolation<Adresse>> violations = validator.validateValue(Adresse.class, "plz", plz);	
		
		if (!violations.isEmpty()) {
			throw new InvalidAdressePlzException(plz, violations);
		}
		LOGGER.debug("ENDE validateAdressePlz");
	}
	
	/**
	 */
	public List<AbstractKunde> findKundenBySeit(Date seit, Locale locale) throws NotFoundException, InvalidKundeSeitException {
		validateKundeSeit(seit, locale);
		
		final List<AbstractKunde> kunden = dao.find(AbstractKunde.FIND_KUNDEN_BY_DATE,
                                               with(AbstractKunde.PARAM_KUNDE_SEIT, seit).parameters(),
                                               AbstractKunde.class);
		return kunden;
	}
	
	private void validateKundeSeit(Date seit, Locale locale) throws InvalidKundeSeitException {
		LOGGER.debug("BEGINN validateKundeSeit: seit={}", seit);
		final Validator validator = validationService.getValidator(locale);
		Set<ConstraintViolation<AbstractKunde>> violations = validator.validateValue(AbstractKunde.class, "seit", seit);	
		
		if (!violations.isEmpty()) {
			throw new InvalidKundeSeitException(seit, violations);
		}
		LOGGER.debug("ENDE validateKundeSeit");
	}
	
	/**
	 */
	public List<AbstractKunde> findPrivatkundenFirmenkunden() throws NotFoundException {
		final List<AbstractKunde> kunden = dao.find(AbstractKunde.FIND_PRIVATKUNDEN_FIRMENKUNDEN,
                                                    AbstractKunde.class);
		return kunden;
	}

	
	/**
	 */
	public AbstractKunde findKundeByUserName(String userName) throws NotFoundException {
		final AbstractKunde kunde = dao.findSingle(AbstractKunde.FIND_KUNDE_BY_USERNAME,
                                                   with(AbstractKunde.PARAM_KUNDE_USERNAME, userName).parameters(),
                                                   AbstractKunde.class);
		return kunde;
	}
	
	/**
	 */
	public List<AbstractKunde> findKundenByNachnameCriteria(String nachname, Locale locale) throws InvalidNachnameException, NotFoundException {
		validateNachname(nachname, locale);
		
		final List<AbstractKunde> kunden = dao.findKundenByNachnameCriteria(nachname);
		return kunden;
	}

	/**
	 */
	public List<AbstractKunde> findKundenMitMinBestMenge(short minMenge) throws NotFoundException {
		final List<AbstractKunde> kunden = dao.findKundenMitMinBestMenge(minMenge);
		return kunden;
	}
	
	/**
	 */
	@RolesAllowed(ROLLE_ADMIN)  // Eigentlich unsinnig; Security soll halt gezeigt werden
	public void deleteKunde(AbstractKunde kunde, Locale locale) throws KundeDeleteBestellungException, InvalidKundeIdException {
		if (kunde == null) {
			return;
		}

		deleteKundeById(kunde.getId(), locale);
	}

	/**
	 */
	@RolesAllowed(ROLLE_ADMIN)  // Eigentlich unsinnig; Security soll halt gezeigt werden
	public void deleteKundeById(Long kundeId, Locale locale) throws KundeDeleteBestellungException, InvalidKundeIdException {
		validateKundeId(kundeId, locale);
		
		AbstractKunde kunde = null;
		try {
			kunde = findKundeMitBestellungenById(kundeId, locale);
		}
		catch (NotFoundException e) {
			// Der Kunde ist bereits geloescht
			return;
		}

		final boolean hasBestellungen = hasBestellungen(kunde);
		if (hasBestellungen) {
			throw new KundeDeleteBestellungException(kunde);
		}

		// Kundendaten loeschen
		dao.delete(kunde, kunde.getId());
	}

	/**
	 */
	@RolesAllowed({ ROLLE_MITARBEITER, ROLLE_ADMIN })  // Eigentlich unsinnig; Security soll halt gezeigt werden
	public AbstractKunde createKunde(AbstractKunde kunde, Locale locale, boolean uiValidated) throws KundeValidationException, KundeDuplikatException, InvalidEmailException {
		if (kunde == null) {
			return kunde;
		}

		// Werden alle Constraints beim Einfuegen gewahrt?
		final Validator validator = validationService.getValidator(locale);
		
		Set<ConstraintViolation<AbstractKunde>> violations = 
			                            uiValidated
			                            ? validator.validate(kunde, PasswordGroup.class)	
			                            : validator.validate(kunde, Default.class, PasswordGroup.class);
		
		if (!violations.isEmpty()) {
			throw new KundeValidationException(kunde, violations);
		}

		// Pruefung, ob ein solcher Kunde schon existiert
		AbstractKunde tmp = null;
		try {
			tmp = findKundeByEmail(kunde.getEmail(), locale);
		}
		catch (NotFoundException e) {
			tmp = null; // Es gibt noch keine Kunden mit diesem Nachnamen: OK
		}
			
		if (tmp != null) {
			throw new KundeDuplikatException(kunde.getId(), kunde.getNachname(), kunde.getVorname(), kunde.getEmail());
		}
		
		// Password verschluesseln
		passwordVerschluesseln(kunde);

		kunde.setId(KEINE_ID);
		dao.create(kunde);
		sendEmail(kunde);
		return kunde;
	}
	
	/**
	 */
	@TransactionAttribute(SUPPORTS)
	public boolean validatePassword(AbstractKunde kunde, String passwort) {
		if (kunde == null) {
			return false;
		}
		
		String verschluesselt = verschluesseln(passwort);
		final boolean result = verschluesselt.equals(kunde.getPassword());
		return result;
	}
	
	/**
	 * Email verschicken
	 */
	@Asynchronous
	@TransactionAttribute(SUPPORTS)
	public Future<Boolean> sendEmail(AbstractKunde kunde) {
		final MimeMessage message = new MimeMessage(mailer);

		// Empfaenger setzen
		final String adressenStr = "Juergen.Zimmermann@HS-Karlsruhe.de";  // durch Komma getrennte Liste
		// RecipientType: TO, CC, BCC
		// Strikte RFC822 Syntax? true/false
		InternetAddress[] adressen;
		try {
			adressen = InternetAddress.parse(adressenStr, false);
		}
		catch (AddressException e) {
			LOGGER.error(e.getMessage());
			return new AsyncResult<Boolean>(Boolean.FALSE);
		}
		
		try {
			message.setRecipients(RecipientType.TO, adressen);

			// Subject setzen
			final String subject = "Neuer Kunde";
			message.setSubject(subject);
			
			// Text setzen mit MIME Type "text/plain"
			final Adresse adr = kunde.getAdresse();
			final String text = kunde.getVorname() + " " + kunde.getNachname() + NEWLINE
			                    + adr.getPlz() + " " + adr.getOrt() + NEWLINE
			                    + adr.getStrasse() + " " + adr.getHausnr() + NEWLINE;
			message.setText(text);
			
			// EMail abschicken
			Transport.send(message);
		}
		catch (MessagingException e) {
			LOGGER.error(e.getMessage());
			return new AsyncResult<Boolean>(Boolean.FALSE);
		}
		
		return new AsyncResult<Boolean>(Boolean.TRUE);
	}


	/**
	 */
	@RolesAllowed({ ROLLE_MITARBEITER, ROLLE_ADMIN })  // Eigentlich unsinnig; Security soll halt gezeigt werden
	public AbstractKunde updateKunde(AbstractKunde kunde, Locale locale, boolean uiValidated, boolean geaendertPassword) throws KundeDuplikatException, KundeValidationException, InvalidEmailException, ConcurrentUpdatedException, ConcurrentDeletedException {
		// Werden alle Constraints beim Modifizieren gewahrt?
		final Validator validator = validationService.getValidator(locale);
		
		final Set<ConstraintViolation<AbstractKunde>> violations = 
			                            uiValidated
			                            ? validator.validate(kunde, PasswordGroup.class, IdGroup.class)	
			                            : validator.validate(kunde, Default.class, PasswordGroup.class, IdGroup.class);
		
		if (!violations.isEmpty()) {
			throw new KundeValidationException(kunde, violations);
		}
		// Pruefung, ob ein solcher Kunde schon existiert
		AbstractKunde tmp = null;
		try {
			tmp = findKundeByEmail(kunde.getEmail(), locale);
		}
		catch (NotFoundException e) {
			tmp = null; // Es gibt noch keine Kunden mit diesem Nachnamen: OK
		}
		if (tmp != null && tmp.getId().longValue() != kunde.getId().longValue()) {
			throw new KundeDuplikatException(kunde.getId(), kunde.getNachname(), kunde.getVorname(), kunde.getEmail());
		}
		
		if (kunde.getClass().equals(Privatkunde.class)) {
			final Privatkunde pk = (Privatkunde) kunde;
			pk.hobbiesToDB();
		}
		
		// Password verschluesseln
		if (geaendertPassword) {
			passwordVerschluesseln(kunde);
		}
		
		kunde = dao.update(kunde, kunde.getId());
		kunde.setPasswordWdh(kunde.getPassword());
		
		return kunde;
	}
	
	/**
	 */
	@RolesAllowed(ROLLE_ADMIN)
	public void addRolle(Long kundeId, String rolle, Locale locale) throws NotFoundException, InvalidKundeIdException {
		validateKundeId(kundeId, locale);
		
		findKundeById(kundeId, locale);

		dao.addRolle(kundeId, rolle);
		refreshRoles(kundeId.toString());
	}

	/**
	 */
	@RolesAllowed(ROLLE_ADMIN)
	public void removeRolle(Long kundeId, String rolle, Locale locale) throws NotFoundException, InvalidKundeIdException {
		validateKundeId(kundeId, locale);
		
		findKundeById(kundeId, locale);

		dao.removeRolle(kundeId, rolle);
		refreshRoles(kundeId.toString());
	}

	/**
	 */
	public Collection<String> getAlleRollen() throws PolicyContextException {
		final Collection<String> rollen = new LinkedList<String>();
		
		// Authentifiziertes Subject ermitteln
		final Subject subject = (Subject) PolicyContext.getContext("javax.security.auth.Subject.container");
		if (subject == null) {
			return null;
		}

		// Gruppe "Roles" ermitteln
		final Set<Principal> principals = subject.getPrincipals(Principal.class);
		for (Principal p : principals) {
			if (!(p instanceof SimpleGroup)) {
				continue;
			}

			final SimpleGroup sg = (SimpleGroup) p;
			if (!"Roles".equals(sg.getName())) {
				continue;
			}
			
			// Rollen ermitteln
			final Enumeration<Principal> members = sg.members();
			while (members.hasMoreElements()) {
				final String rolle = members.nextElement().toString();
				if (rolle != null) {
					rollen.add(rolle);
				}
			}
		}
		return rollen;
	}

	
	/**
	 */
	public AbstractKunde neuesBild(Long kundeId, byte[] bildBytes, String contentType, String extension, Locale locale) throws NotFoundException, InvalidKundeIdException, BildException {
		validateKundeId(kundeId, locale);
		
		// Kunde ermitteln und aktualisieren
		AbstractKunde k = findKundeById(kundeId, locale);
		k = dao.updateKundeMitBild(k, bildBytes, contentType, extension);

		return k;
	}
	
	/**
	 */
	public AbstractKunde neuesBild(Long kundeId, File bildTmpFile, String contentType, String extension, Locale locale) throws NotFoundException, InvalidKundeIdException, BildException {
		validateKundeId(kundeId, locale);
		
		// Kunde ermitteln und aktualisieren
		AbstractKunde k = findKundeById(kundeId, locale);
		k = dao.updateKundeMitBild(k, bildTmpFile, contentType, extension);

		return k;
	}
	
	/**
	 * Methode, um zu zeigen, wie ein explizites Rollback funktioniert.
	 */
	@Deprecated
	public void updateKundeMitRollback(AbstractKunde kunde) throws ConcurrentUpdatedException, ConcurrentDeletedException {
		dao.update(kunde, kunde.getId());
		sessionCtx.setRollbackOnly();
	}
	
	/**
	 */
	public Betreuer findBetreuer(Long id, Locale locale) throws NotFoundException, InvalidBetreuerIdException {
		validateBetreuerId(id, locale);
		
		final Betreuer betreuer = dao.find(Betreuer.class, id);
		return betreuer;
	}
	
	/**
	 */
	private void validateBetreuerId(Long id, Locale locale) throws InvalidBetreuerIdException {
		LOGGER.debug("BEGINN validateBetreuerId: id={}", id);
		final Validator validator = validationService.getValidator(locale);
		final Set<ConstraintViolation<Betreuer>> violations = validator.validateValue(Betreuer.class, "id", id, IdGroup.class);	
		
		if (!violations.isEmpty()) {
			throw new InvalidBetreuerIdException(id, violations);
		}
		
		LOGGER.debug("ENDE validateBetreuerId");
	}
	
	/**
	 */
	public List<Wartungsvertrag> findWartungsvertraege(Long kundeId, Locale locale)
	                             throws NotFoundException, InvalidKundeIdException {
		validateKundeId(kundeId, locale);
		
		final List<Wartungsvertrag> wartungsvertraege = dao.find(Wartungsvertrag.FIND_WARTUNGSVERTRAEGE_BY_KUNDE_ID,
                                                                 with(Wartungsvertrag.PARAM_KUNDE_ID, kundeId).parameters(),
                                                                 Wartungsvertrag.class);
		return wartungsvertraege;
	}

	
	/**
	 */
	public Wartungsvertrag createWartungsvertrag(Wartungsvertrag wartungsvertrag) throws NotFoundException {
		if (wartungsvertrag == null) {
			return null;
		}
		
		wartungsvertrag = dao.create(wartungsvertrag);
		return wartungsvertrag;
	}
	
	/**
	 */
	private void validateKundeId(Long kundeId, Locale locale) throws InvalidKundeIdException {
		LOGGER.debug("BEGINN validateKundeId: kundeId={}", kundeId);
		final Validator validator = validationService.getValidator(locale);
		Set<ConstraintViolation<AbstractKunde>> violations = validator.validateValue(AbstractKunde.class, "id", kundeId, IdGroup.class);	
		
		if (!violations.isEmpty()) {
			throw new InvalidKundeIdException(kundeId, violations);
		}
		
		LOGGER.debug("ENDE validateKundeId");
	}
	
	/**
	 */
	private void validateNachname(String nachname, Locale locale) throws InvalidNachnameException {
		LOGGER.debug("BEGINN validateNachname: nachname={}", nachname);
		final Validator validator = validationService.getValidator(locale);
		
		final Set<ConstraintViolation<AbstractKunde>> violations = validator.validateValue(AbstractKunde.class, "nachname", nachname);	

		if (!violations.isEmpty()) {
			throw new InvalidNachnameException(nachname, violations);
		}
		
		LOGGER.debug("ENDE validateNachname");
	}

	
	/**
	 */
	private boolean hasBestellungen(AbstractKunde kunde) {
		LOGGER.debug("BEGINN hasBestellungen: {}", kunde);
		
		boolean result = false;
		
		// Gibt es den Kunden und hat er mehr als eine Bestellung?
		// Bestellungen nachladen wegen Hibernate-Caching
		if (kunde != null && kunde.getBestellungen() != null && !kunde.getBestellungen().isEmpty()) {
			result = true;
		}
		
		LOGGER.debug("ENDE hasBestellungen: {}", result);
		return result;
	}

	
	private static void passwordVerschluesseln(AbstractKunde kunde) {
		LOGGER.debug("BEGINN passwordVerschluesseln: kunde={}", kunde);

		final String unverschluesselt = kunde.getPassword();
		final String verschluesselt = verschluesseln(unverschluesselt);
		kunde.setPassword(verschluesselt);
		kunde.setPasswordWdh(verschluesselt);

		LOGGER.debug("ENDE passwordVerschluesseln: {}", verschluesselt);
	}

}
