package de.swe2.kundenverwaltung.service;

import static de.swe2.util.AbstractDao.QueryParameter.with;
import static de.swe2.util.EjbConstants.BLOB_PATH_WEB;
import static de.swe2.util.EjbConstants.NAME_WEB_PROJEKT;
import static de.swe2.util.EjbConstants.ROLLE_TABELLE;
import static javax.ejb.TransactionAttributeType.MANDATORY;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.List;

import javax.ejb.Stateless;
import javax.ejb.TransactionAttribute;
import javax.persistence.EntityExistsException;
import javax.persistence.Query;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Join;
import javax.persistence.criteria.JoinType;
import javax.persistence.criteria.Path;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.jboss.classloader.spi.base.BaseClassLoader;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.swe2.bestellverwaltung.domain.Bestellposition;
import de.swe2.bestellverwaltung.domain.Bestellung;
import de.swe2.kundenverwaltung.domain.AbstractKunde;
import de.swe2.util.AbstractDao;
import de.swe2.util.NotFoundException;

@Stateless
@TransactionAttribute(MANDATORY)
public class KundenverwaltungDao extends AbstractDao {
	private static final long serialVersionUID = -4911683428419830303L;

	private static final Logger LOGGER = LoggerFactory.getLogger(KundenverwaltungDao.class);
	
	private static final String VFSFILE_PROTOCOL = "vfsfile:/";
	private static final String WEB_SUFFIX = ".war";
	
	private static final long MAX_BYTES_WRITE = (64 * 1024 * 1024) - (32 * 1024);
	private static final String BLOBPATH = getBlobPath();
	
	private static String getBlobPath() {
		
		final ClassLoader cl = KundenverwaltungDao.class.getClassLoader();
		if (cl.getClass() != BaseClassLoader.class) {
			LOGGER.warn("JBossAS ist nicht die Laufzeitumgebung");
			return null;
		}
		
		final BaseClassLoader classloader = (BaseClassLoader) cl;
		final String urlString = classloader.getName();
		
		// Pfad fuer BLOB aufbauen aus z.B. 
		// vfsfile:/C:/.../.metadata/.plugins/org.jboss.ide.eclipse.as.core/JBoss_6.0_Runtime_Server<Zahl>/deploy/swe2.ear/
		// Ergebnis:
		// C:/.../.metadata/.plugins/org.jboss.ide.eclipse.as.core/JBoss_6.0_Runtime_Server<Zahl>/deploy/swe2.ear/swe2Web.war/resources/db/
		
		// "vfsfile:/" loeschen und <MyProjectWeb>.war/resources/db/ anhaengen
		final StringBuilder sb = new StringBuilder(urlString);
		sb.delete(0, VFSFILE_PROTOCOL.length());
		sb.append('/');
		sb.append(NAME_WEB_PROJEKT);
		sb.append(WEB_SUFFIX);
		sb.append(BLOB_PATH_WEB);
		
		final String blobPath = sb.toString();
		LOGGER.info("Pfad fuer BLOBs: " + blobPath);
		
		return blobPath;
	}

	public List<AbstractKunde> findKundenMitBestellungenUndLieferungen(String nachname)
                               throws NotFoundException {
		final List<AbstractKunde> kunden = find(AbstractKunde.FIND_KUNDEN_BY_NACHNAME_FETCH_BESTELLUNGEN,
                                                with(AbstractKunde.PARAM_KUNDE_NACHNAME, nachname).parameters(),
                                                AbstractKunde.class);
		// 2. Fetch-Join als Criteria Query:
		//       SELECT b
		//       FROM   Bestellung b
		//              LEFT JOIN FETCH b.lieferungen
		//       WHERE  b.id = ? OR b.id = ? ...

		final CriteriaBuilder builder = em.getCriteriaBuilder();
		final CriteriaQuery<Bestellung> criteriaQuery = builder.createQuery(Bestellung.class);
		final Root<Bestellung> b = criteriaQuery.from(Bestellung.class);
		b.fetch("lieferungen", JoinType.LEFT);
		
		final Path<Long> idPath = b.get("id");
		
		final List<Predicate> predList = new ArrayList<Predicate>();
		for (AbstractKunde k : kunden) {
			final List<Bestellung> bestellungen = k.getBestellungen();
			for (Bestellung best : bestellungen) {
				final Predicate equal = builder.equal(idPath, best.getId());
				predList.add(equal);
			}
		}
			
		if (!predList.isEmpty()) {
			final Predicate[] predArray = new Predicate[predList.size()];
			final Predicate pred = builder.or(predList.toArray(predArray));
			criteriaQuery.where(pred).distinct(true);
			
			final TypedQuery<Bestellung> queryObj = em.createQuery(criteriaQuery);
			queryObj.getResultList();
			
			// Ergebnis der Query nicht relevant, da nur Fetching benoetigt wird
		}
	
		return kunden;
	}
	

	/**
	 */
	public List<AbstractKunde> findKundenByNachnameCriteria(String nachname) throws NotFoundException {
		final CriteriaBuilder builder = em.getCriteriaBuilder();
		final CriteriaQuery<AbstractKunde> criteriaQuery = builder.createQuery(AbstractKunde.class);
		final Root<AbstractKunde> k = criteriaQuery.from(AbstractKunde.class);

		//final Path<String> nachnamePath = k.get(AbstractKunde_.nachname);
		final Path<String> nachnamePath = k.get("nachname");
		
		final Predicate pred = builder.equal(nachnamePath, nachname);
		criteriaQuery.where(pred);

		final TypedQuery<AbstractKunde> query = em.createQuery(criteriaQuery);
		final List<AbstractKunde> kunden = query.getResultList();
		
		if (kunden.isEmpty()) {
			throw new NotFoundException("Kein Treffer fuer die Criteria Query: Suche Kunden mit Nachname " + nachname, AbstractKunde.class);
		}
		
		return kunden;
	}
	
	/**
	 */
	public List<AbstractKunde> findKundenMitMinBestMenge(short minMenge) throws NotFoundException {
		final CriteriaBuilder builder = em.getCriteriaBuilder();
		final CriteriaQuery<AbstractKunde> criteriaQuery  = builder.createQuery(AbstractKunde.class);
		final Root<AbstractKunde> k = criteriaQuery.from(AbstractKunde.class);

		//final Join<AbstractKunde, Bestellung> b = k.join(AbstractKunde_.bestellungen);
		//final Join<Bestellung, Bestellposition> bp = b.join(Bestellung_.bestellpositionen);
		//criteriaQuery.where(builder.gt(bp.<Short>get(Bestellposition_.anzahl), minMenge))
		//             .distinct(true);
		final Join<AbstractKunde, Bestellung> b = k.join("bestellungen");
		final Join<Bestellung, Bestellposition> bp = b.join("bestellpositionen");
		criteriaQuery.where(builder.gt(bp.<Short>get("anzahl"), minMenge))
		             .distinct(true);

		final TypedQuery<AbstractKunde> query = em.createQuery(criteriaQuery);
		final List<AbstractKunde> kunden = query.getResultList();
		
		if (kunden.isEmpty()) {
			throw new NotFoundException("Kein Treffer fuer die Criteria Query: Suche Kunden mit mind. Bestellmenge " + minMenge, AbstractKunde.class);
		}
		return kunden;
	}
	
	/**
	 */
	public AbstractKunde reloadKunde(AbstractKunde kunde) throws NotFoundException {
		if (em.contains(kunde)) {
			LOGGER.trace("Kein Reload von " + kunde);
			return kunde;
		}
		
		LOGGER.trace("Reload: " + kunde);
		final Long id = kunde.getId();
		kunde = em.find(AbstractKunde.class, id);
		if (kunde == null) {
			throw new NotFoundException("Der Kunde mit ID=" + id + " existiert nicht mehr", AbstractKunde.class);
		}

		return kunde;
	}
	
	/**
	 */
	public void addRolle(Long kundeId, String rolle) {
		// Neue Rolle mittels SQL in die Tabelle einfuegen
		final String queryStr = "INSERT INTO " + ROLLE_TABELLE + " VALUES(" + kundeId + ", '" + rolle + "')";
		LOGGER.trace(queryStr);
		final Query query = em.createNativeQuery(queryStr);
		try {
			query.executeUpdate();
		}
		catch (EntityExistsException e) {
			LOGGER.warn("Der Kunde mit ID " + kundeId + " hat bereits die Rolle " + rolle);
			return;
		}
	}

	/**
	 */
	public void removeRolle(Long kundeId, String rolle) {
		final String queryStr = "DELETE FROM " + ROLLE_TABELLE + " WHERE kunde_fk = " + kundeId + " AND role = '" + rolle + "'";
		LOGGER.trace(queryStr);
		final Query query = em.createNativeQuery(queryStr);
		query.executeUpdate();
	}

	/**
	 */
	public AbstractKunde updateKundeMitBild(AbstractKunde k, File tmpFile, String contentType, String extension) throws BildException {
		if (tmpFile == null) {
			return null;
		}
		
		// Dateiname mit Extension ermitteln
		final String neuerDateiname = k.getClass().getSimpleName() + "_" + k.getId() + "_" + k.getVersion() + '.' + extension;
		LOGGER.trace(neuerDateiname);
		
		// temporaere Datei umkopieren
		FileChannel inChannel = null;
		FileChannel outChannel = null;
		try {
			inChannel = new FileInputStream(tmpFile).getChannel();
			final File neueDatei = new File(BLOBPATH + File.separator + neuerDateiname);
			outChannel = new FileOutputStream(neueDatei).getChannel();
			final long dateiGroesse = inChannel.size();
			for (long position = 0; position < dateiGroesse; ) {
				position += inChannel.transferTo(position, MAX_BYTES_WRITE, outChannel);
			}
		}
		catch (IOException e) {
			throw new RuntimeException(e);
		}
		finally {
			try {
				if (inChannel != null) {
					inChannel.close();
				}
				if (outChannel != null) {
					outChannel.close();
				}
			}
			catch (IOException e) {
				throw new RuntimeException(e);
			}
		}

		final String alterDateiname = k.getDateiname();
		k.setDateiname(neuerDateiname);
		k.setDateiContentType(contentType);
		
		// Alte Datei ggf. loeschen
		if (alterDateiname != null && !alterDateiname.equals("")) {
			final File file = new File(BLOBPATH + File.separator + alterDateiname);
			try {
				file.delete();
			}
			catch (SecurityException e) {
				throw new IllegalStateException(e);
			}
		}

		return k;
	}
	

	/**
	 */
	public AbstractKunde updateKundeMitBild(AbstractKunde k, byte[] bildBytes, String contentType, String extension) throws BildException {
		// Dateiname mit Extension ermitteln
		final String neuerDateiname = k.getClass().getSimpleName() + "_" + k.getId() + "_" + k.getVersion() + '.' + extension;
		
		// Bild zum gefundenen Kunden abspeichern
		final ByteBuffer byteBuffer = ByteBuffer.wrap(bildBytes);
		FileChannel outChannel = null;
		try {
			final File neueDatei = new File(BLOBPATH + File.separator + neuerDateiname);
			outChannel = new FileOutputStream(neueDatei).getChannel();
			outChannel.write(byteBuffer);
		}
		catch (IOException e) {
			throw new RuntimeException(e);			
		}
		finally {
			try {
				if (outChannel != null) {
					outChannel.close();
				}
			}
			catch (IOException e) {
				throw new RuntimeException(e);
			}
		}

		final String alterDateiname = k.getDateiname();
		k.setDateiname(neuerDateiname);
		k.setDateiContentType(contentType);
		
		// Alte Datei ggf. loeschen
		if (alterDateiname != null && !alterDateiname.equals("")) {
			final File alteDatei = new File(BLOBPATH + File.separator + alterDateiname);
			try {
				alteDatei.delete();
			}
			catch (SecurityException e) {
				throw new IllegalStateException(e);
			}
		}

		return k;
	}
	
//	Mit NIO2 von OpenJDK 7
//	/*
//	 */
//	@TransactionAttribute(MANDATORY)
//	public Kunde updateKundeMitBild(Kunde k, File tmpFile, String contentType, String extension) throws BildException {
//		if (tmpFile == null)
//			return null;
//		
//		// Dateiname mit Extension ermitteln
//		final String neuerDateiname = k.getClass().getSimpleName() + "_" + k.getId() + "_" + k.getVersion() + '.' + extension;
//		if (TRACE) LOGGER.trace(neuerDateiname);
//		
//		// Bild zum gefundenen Kunden abspeichern mit NIO2 aus OpenJDK 7
//		final Path tmpPath = Paths.get(tmpFile.getAbsolutePath());
//		final Path blobPath = Paths.get(BLOBPATH + neuerDateiname);
//		
//		try {
//			tmpPath.moveTo(blobPath);
//		}
//		catch (IOException e) {
//			throw new BildException(k.getId(), neuerDateiname, e);
//		}
//		
//		final String alterDateiname = k.getDateiname();
//		k.setDateiname(neuerDateiname);
//		k.setDateiContentType(contentType);
//		
//		// Alte Datei ggf. loeschen
//		if (alterDateiname != null && !alterDateiname.equals("")) {
//			final Path alteDatei = Paths.get(BLOBPATH + alterDateiname);
//			try {
//				alteDatei.delete();
//			}
//			catch (IOException e) {
//				throw new BildException(k.getId(), alterDateiname, e);
//			}
//		}
//
//		return k;
//	}


//	Mit NIO2 von OpenJDK 7
//	/*
//	 */
//	@TransactionAttribute(MANDATORY)
//	public Kunde updateKundeMitBild(Kunde k, byte[] bildBytes, String contentType, String extension) throws BildException {
//		// Dateiname mit Extension ermitteln
//		final String neuerDateiname = k.getClass().getSimpleName() + "_" + k.getId() + "_" + k.getVersion() + '.' + extension;
//		
//		// Bild zum gefundenen Kunden abspeichern mit NIO2 aus OpenJDK 7
//		final Path blobPath = Paths.get(BLOBPATH + neuerDateiname);
//		OutputStream out = null;
//		try {
//			out = blobPath.newOutputStream(StandardOpenOption.CREATE, StandardOpenOption.WRITE);
//			out.write(bildBytes);
//		}
//		catch (IOException e) {
//			throw new BildException(k.getId(), neuerDateiname, e);
//		}
//		finally {
//			try {
//				if (out != null)
//					out.close();
//			}
//			catch (IOException e) {
//				throw new BildException(k.getId(), neuerDateiname, e);
//			}			
//		}
//		final String alterDateiname = k.getDateiname();
//		k.setDateiname(neuerDateiname);
//		k.setDateiContentType(contentType);
//		
//		// Alte Datei ggf. loeschen
//		if (alterDateiname != null && !alterDateiname.equals("")) {
//			final Path alteDatei = Paths.get(BLOBPATH + alterDateiname);
//			try {
//				alteDatei.delete();
//			}
//			catch (IOException e) {
//				throw new BildException(k.getId(), alterDateiname, e);
//			}
//		}
//
//		return k;
//	}
}
