package de.swe2.util;


import java.io.Serializable;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.ResourceBundle;

import javax.enterprise.context.ApplicationScoped;
import javax.faces.application.FacesMessage;
import javax.faces.application.FacesMessage.Severity;
import javax.faces.context.FacesContext;
import javax.inject.Named;
import javax.validation.ConstraintViolation;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@Named("messages")
@ApplicationScoped
@Log
public class MessagesMgdBean implements Serializable {
	private static final long serialVersionUID = -8733240117333288841L;
	private static final Logger LOGGER = LoggerFactory.getLogger(MessagesMgdBean.class);
			
	private static final String DEFAULT_LANG = Locale.GERMAN.getLanguage();
	private static final List<String> languages = Arrays.asList(DEFAULT_LANG,
			                                                    Locale.ENGLISH.getLanguage());

	public enum MessagesType implements Serializable {
		SWE2("de.swe2.messages"),
		KUNDENVERWALTUNG("de.swe2.kundenverwaltung.ui.messages"),
		BESTELLVERWALTUNG("de.swe2.bestellverwaltung.ui.messages"),
		ARTIKELVERWALTUNG("de.swe2.artikelverwaltung.ui.messages");
		
		private String value;
		
		MessagesType(String value) {
			this.value = value;
		}
		
		public String value() {
			return value;
		}
		
	    public static MessagesType fromValue(String value) {
	        for (MessagesType mt: MessagesType.values()) {
	            if (mt.value.equals(value)) {
	                return mt;
	            }
	        }
	        throw new IllegalArgumentException(value);
	    }
	}
	
	private Map<MessagesType, Map<String, ResourceBundle>> bundles;  // z.B. "de" als Schluessel auch fuer de_DE
	
	public MessagesMgdBean() {
		super();
		
		bundles = new HashMap<MessagesType, Map<String, ResourceBundle>>();
		for (MessagesType messagesType : MessagesType.values()) {
			final Map<String, ResourceBundle> bundleMap = new HashMap<String, ResourceBundle>();
			bundles.put(messagesType, bundleMap);

			for (String lang : languages) {
				ResourceBundle bundle = ResourceBundle.getBundle(messagesType.value, new Locale(lang));
				bundleMap.put(lang, bundle);				
			}
		}
	}

	
	/**
	 * F&uuml;r das Exception-Handling bei Managed Beans
	 */
	public void createFacesMessage(MessagesType messagesType, String msgKey, String clientId, Severity severity, Object... args) {
		final FacesContext ctx = FacesContext.getCurrentInstance();
		final Locale locale = ctx.getViewRoot().getLocale();
		
		ResourceBundle bundle = bundles.get(messagesType).get(locale.getLanguage());
		if (bundle == null) {
			// Keine Texte zu aktuellen Sprache gefunden: Default-Sprache verwenden
			bundle = bundles.get(messagesType).get(DEFAULT_LANG);
		}
		final String msgPattern = bundle.getString(msgKey);
		final MessageFormat formatter = new MessageFormat(msgPattern);
		final String msg = formatter.format(args);
		
		final FacesMessage facesMsg = new FacesMessage(severity, msg, null);
		ctx.addMessage(clientId, facesMsg);
	}


	/**
	 * F&uuml;r das Exception-Handling bei Managed Beans, falls die Meldung durch Bean Validation
	 * bereits gegeben ist, z.B. bei Validierung in Suchmasken oder bei @assertTrue
	 */
	public <T> void createFacesMessage(Collection<ConstraintViolation<T>> violations, String clientId, Severity severity) {
		final FacesContext ctx = FacesContext.getCurrentInstance();
		for (ConstraintViolation<T> v: violations) {
			LOGGER.trace("{}", v);
			final FacesMessage facesMsg = new FacesMessage(severity, v.getMessage(), null);
			ctx.addMessage(clientId, facesMsg);
		}
	}

	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder("MessagesMgdBean [bundles={");
		for (MessagesType mt : bundles.keySet()) {
			sb.append(mt.value);
			sb.append(" : {");
			final Map<String, ResourceBundle> bundleMap = bundles.get(mt);
			for (String localeStr : bundleMap.keySet()) {
				sb.append(localeStr);
				sb.append(", ");
			}
			final int length = sb.length();
			sb.delete(length - 2, length - 1);
			sb.append("}, ");
		}
		
		final int length = sb.length();
		sb.delete(length - 2, length - 1);
		sb.append("}");
		
		return sb.toString();
	}
}
